from edc_constants.constants import DM
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import (
    raise_if_clinical_review_does_not_exist,
    raise_if_initial_review_does_not_exist,
)
from edc_form_validators import FormValidator
from edc_visit_schedule.utils import raise_if_baseline

from .glucose_form_validator_mixin import GlucoseFormValidatorMixin


class GlucoseFormValidator(
    GlucoseFormValidatorMixin,
    CrfFormValidatorMixin,
    FormValidator,
):
    """Declared as an example of the clean method to use with
    the mixin.
    """

    required_at_baseline = True
    require_diagnosis = False
    require_clinical_review_crf = True
    prefix = "glucose"

    def clean(self):
        if self.cleaned_data.get("subject_visit"):
            if not self.required_at_baseline:
                raise_if_baseline(self.cleaned_data.get("subject_visit"))
            self.raise_if_clinical_review_does_not_exist()
            if self.require_diagnosis:
                raise_if_initial_review_does_not_exist(
                    self.cleaned_data.get("subject_visit"), DM
                )
            self.validate_glucose_test()

    def raise_if_clinical_review_does_not_exist(self) -> None:
        if self.require_clinical_review_crf:
            return raise_if_clinical_review_does_not_exist(
                self.cleaned_data.get("subject_visit")
            )
        return None



from django.db import models
from django.utils.html import format_html
from edc_constants.constants import NOT_APPLICABLE
from edc_lab.choices import GLUCOSE_UNITS_NA, RESULT_QUANTIFIER
from edc_lab.constants import EQ
from edc_model.validators import datetime_not_future

from ..constants import GLUCOSE_HIGH_READING


def ogtt_model_mixin_factory(utest_id: str, **kwargs):
    class AbstractModel(models.Model):
        class Meta:
            abstract = True

    opts = {
        f"{utest_id}_base_datetime": models.DateTimeField(
            verbose_name="Date/time oral glucose solution given",
            validators=[datetime_not_future],
            null=True,
            blank=True,
            help_text="(glucose solution given)",
        ),
        f"{utest_id}_value": models.DecimalField(
            verbose_name=format_html(
                "Blood glucose measure 2hrs <u>after</u> oral glucose solution given"
            ),
            max_digits=8,
            decimal_places=2,
            null=True,
            blank=True,
            help_text=f"A `HIGH` reading may be entered as {GLUCOSE_HIGH_READING}",
        ),
        f"{utest_id}_quantifier": models.CharField(
            max_length=10,
            choices=RESULT_QUANTIFIER,
            default=EQ,
        ),
        f"{utest_id}_units": models.CharField(
            verbose_name="Units (Blood glucose 2hrs after...)",
            max_length=15,
            default=NOT_APPLICABLE,
            choices=GLUCOSE_UNITS_NA,
            blank=False,
        ),
        f"{utest_id}_datetime": models.DateTimeField(
            verbose_name=format_html(
                "Date/time blood glucose measured 2hrs <u>after</u> "
                "oral glucose solution given"
            ),
            validators=[datetime_not_future],
            blank=True,
            null=True,
            help_text="(2 hours after glucose solution given)",
        ),
    }

    opts.update(**kwargs)

    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel


class OgttModelMixin(ogtt_model_mixin_factory("ogtt"), models.Model):
    """A model mixin of fields for the OGTT"""

    class Meta:
        abstract = True



from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_appointment.tests.helper import Helper as BaseHelper
from edc_consent import site_consents
from edc_constants.constants import NOT_APPLICABLE, YES
from edc_lab.constants import EQ
from edc_reportable import MILLIMOLES_PER_LITER
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.models import SubjectVisit

from edc_glucose.form_validators import GlucoseFormValidator

from ..consents import consent_v1
from ..models import SubjectScreening
from ..visit_schedules import visit_schedule


class Helper(BaseHelper):
    @property
    def screening_model_cls(self):
        return SubjectScreening


@override_settings(SITE_ID=10)
class TestGlucose(TestCase):
    helper_cls = Helper

    def setUp(self):
        SubjectScreening.objects.create(
            screening_identifier="ABCD",
            screening_datetime=get_utcnow(),
            subject_identifier="12345",
        )
        self.subject_identifier = "12345"
        site_consents.registry = {}
        site_consents.register(consent_v1)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
        )
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
        )

        appointment_baseline = Appointment.objects.all().order_by(
            "timepoint", "visit_code_sequence"
        )[0]

        self.subject_visit_baseline = SubjectVisit.objects.create(
            appointment=appointment_baseline,
            subject_identifier=self.subject_identifier,
            visit_code=1000,
            visit_code_sequence=0,
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
            reason=SCHEDULED,
        )
        appointment_followup = Appointment.objects.all().order_by(
            "timepoint", "visit_code_sequence"
        )[1]
        self.subject_visit_followup = SubjectVisit.objects.create(
            appointment=appointment_followup,
            subject_identifier=self.subject_identifier,
            visit_code="2000",
            visit_code_sequence=0,
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
            reason=SCHEDULED,
        )

    def test_glucose_result(self):
        class MyGlucoseFormValidator(GlucoseFormValidator):
            def clean(self):
                self.validate_glucose_test()

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=NOT_APPLICABLE,
            glucose_units=NOT_APPLICABLE,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_fasting", cm.exception.error_dict)

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=YES,
            glucose_units=NOT_APPLICABLE,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_fasting_duration_str", cm.exception.error_dict)

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=YES,
            glucose_fasting_duration_str="12h",
            glucose_units=NOT_APPLICABLE,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_date", cm.exception.error_dict)

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=YES,
            glucose_fasting_duration_str="12h",
            glucose_date=get_utcnow().date,
            glucose_units=NOT_APPLICABLE,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_value", cm.exception.error_dict)

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=YES,
            glucose_fasting_duration_str="12h",
            glucose_date=get_utcnow().date,
            glucose_value=5.3,
            glucose_units=MILLIMOLES_PER_LITER,
            glucose_quantifier=None,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_quantifier", cm.exception.error_dict)

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=YES,
            glucose_fasting_duration_str="12h",
            glucose_date=get_utcnow().date,
            glucose_value=5.3,
            glucose_units=NOT_APPLICABLE,
            glucose_quantifier=EQ,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_units", cm.exception.error_dict)

        cleaned_data = dict(
            subject_visit=self.subject_visit_baseline,
            glucose_performed=YES,
            glucose_fasting=YES,
            glucose_fasting_duration_str="12h",
            glucose_date=get_utcnow().date,
            glucose_value=5.3,
            glucose_units=MILLIMOLES_PER_LITER,
            glucose_quantifier=EQ,
        )
        form_validator = MyGlucoseFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised")



from .fasting_form_validator import FastingFormValidatorMixin
from .fbg_form_validator_mixin import FbgFormValidatorMixin
from .fbg_ogtt_form_validator_mixin import FbgOgttFormValidatorMixin
from .glucose_form_validator import GlucoseFormValidator
from .glucose_form_validator_mixin import GlucoseFormValidatorMixin
from .ogtt_form_validator_mixin import OgttFormValidatorMixin
